//----------------------------------------------------------------------------
// C main line
//----------------------------------------------------------------------------

#include <m8c.h>        // part specific constants and macros
#include "PSoCAPI.h"    // PSoC API definitions for all User Modules
#include "globdefs.h"
#include "utils.h"
#include "time.h"
#include "batcheck.h"
#include "globalparams.h"



BYTE state = POWER_ON, new_state = BAT_POWER;
static BYTE led_state = LED_OFF;

static INT prs_val, act_val;
static BOOL limit_val;

static INT vload, iload;
static INT vload_filt;

static BYTE charge_complete = FALSE;

static BYTE sleep_wakeup = FALSE;
static volatile BYTE send_debug_info = FALSE;
static volatile BYTE button_released = FALSE;
static volatile BYTE button_pressed =FALSE;

static INT bat_temp = 0;
static INT duty = 0;

#if DEBUG
BYTE error_code = NO_ERROR;
#endif


#define SLEEP_TIMEOUT 16

#define LPF_SHIFT 2

#define CHECK_MAIN_SUPPLY() CHECKBIT(MAINS_POWER_PORT, MAINS_POWER_BIT)
#define REGULATOR_RESET() duty = 0;

#define SET_SUPPLY_VOLTAGE_LOW() VLT_CR &= ~0x07;
#define SET_SUPPLY_VOLTAGE_HIGH() VLT_CR |= 0x04;


#if DEBUG
void SendDebugData(void)
{ 
  Tx232_CWrite("S",1);
  Tx232_PutSHexByte(state);
  Tx232_PutSHexByte(error_code);
  Tx232_PutSHexInt(vload);
  Tx232_PutSHexInt(iload);
  Tx232_PutSHexInt(bat_temp);
  Tx232_CWrite("P",1);
}
#endif


#pragma interrupt_handler sleep_timer_handler;
void sleep_timer_handler(void)
{  
  SETBIT(BUTTON_PORT, BUTTON_BIT);
  if (CHECKBIT(BUTTON_PORT, BUTTON_BIT)) button_released = TRUE;
    else 
  {
    if (button_released) button_pressed = TRUE;
    button_released = FALSE;
  }  
  
  CLEARBIT(BUTTON_PORT, BUTTON_BIT);
   
  if (LED_ON == led_state) CLEARBIT(LED_PORT, LED_BIT);
  else if (led_state == LED_BLINK) LED_PORT ^= BIT(LED_BIT); 
  else SETBIT(LED_PORT, LED_BIT);
  
#if DEBUG   
  send_debug_info = TRUE;
#endif  
}

void delay(unsigned char cycles)
{
  while (cycles--);
}

BOOL sleep(void)
{
  BOOL temp = FALSE;
  
  ABF_CR0 = 0;
  ARF_CR &= ~ANALOG_POWER_MASK;  
  
  M8C_DisableIntMask(INT_MSK0, INT_MSK0_VC3);
  M8C_EnableIntMask(INT_MSK0, INT_MSK0_GPIO);
  
  SETBIT(BUTTON_PORT, BUTTON_BIT);
   
  OSC_CR0 |= OSC_CR0_SLEEP_1Hz;
  
  M8C_DisableGInt;
  
  M8C_ClearWDTAndSleep;
  INT_VC = 0;
  
  M8C_Sleep;
  
  if(CHECKBIT(INT_CLR0, 6)) return TRUE;

  OSC_CR0 &= ~0x08;
  M8C_EnableIntMask(INT_MSK0, INT_MSK0_VC3);
  M8C_DisableIntMask(INT_MSK0, INT_MSK0_GPIO);
  
  while(!CHECKBIT(INT_CLR0,6));
   
  ABF_CR0 = ANALOG_IO_CONTROL;
  ARF_CR |= ANALOG_POWER;
      
  M8C_EnableGInt;
    
  return FALSE;  
}

INT get_adc_sample(void)
{
   ADCINC_GetSamples(1);
   while(!ADCINC_fIsDataAvailable());
   return ADCINC_iClearFlagGetData();
}

BYTE check_button_pressing(void)
{
  BYTE temp;
  
  temp = button_pressed;
  button_pressed = 0; 
   
  return temp;    
}



void measure_load_voltage(void)
{
  BYTE repeats = 2;
  
  vload = 0;
  
  ADCINC_AtoDcr1 = 0xA0;
  
  while (repeats--)
  {
    AMX_IN = 0x03;
    vload += get_adc_sample();
    
    AMX_IN = 0x02;
    vload -= get_adc_sample();
  }
   
  vload >>= 1;  
  
  vload_filt  -= (vload_filt >> LPF_SHIFT) - vload;
}

void measure_load_current(void)
{  
  BYTE repeats = 2;   
  iload = 0;
  
  ADCINC_AtoDcr1 = 0x60;
  
  while(repeats--)
  {
    AMX_IN = 0x08;    
    delay(FILTER_SETUP);  
    iload += get_adc_sample();
  
    AMX_IN = 0x0C;
    delay(FILTER_SETUP);
    iload -= get_adc_sample();
  }  
  
  iload >>= 1;
}

void measure_bat_temperature(void)
{
  AMX_IN = 0;
  ADCINC_AtoDcr1 = 0xA0;
  
  PGA_GAIN_CR2 |= 0x1C;
  bat_temp = get_adc_sample();
  
  PGA_GAIN_CR2 &= ~0x08;
  bat_temp -= get_adc_sample();  
}

BOOL check_temp_range(void)
{ 
  static BOOL old_st = TRUE;  
  BYTE new_st;
  
  if ((bat_temp > TEMP_LOW_OFF) || (bat_temp < TEMP_HIGH_OFF)) new_st = FALSE;
    else 
  if ((bat_temp > TEMP_HIGH_ON) && (bat_temp < TEMP_LOW_ON)) new_st = TRUE;
    else
  new_st = old_st;  
    
  old_st = new_st;
  
  return new_st;
}

void pfm_regulate(void)
{  
   static BYTE step;
   static INT diff, abs_diff;
   
   diff = prs_val - act_val;
   abs_diff = (diff > 0)?(diff):(-diff);
   
   if (abs_diff > DIF1) step = STEP1;
     else 
   if (abs_diff > DIF2) step = STEP2;
     else 
   if (abs_diff > DIF3) step = STEP3;
     else
   step = 1;

     
   if ((diff > DIF_DELTA) && (0 == limit_val)) duty += step;
   if ((diff < DIF_DELTA) || (0 != limit_val)) duty -= step;
   								 
   CHECKRANGE(duty, DUTY_MIN, DUTY_MAX);	
   
   PRS8_SEED_REG = (BYTE)duty;						 
}   

void state_machine(void)
{
  new_state = state;

  if(CHECK_MAIN_SUPPLY()) {new_state &= 0x07; new_state |= 0x02;} else new_state &= ~0x06;  
  if (check_button_pressing()) {new_state &= 0x03; new_state ^= 0x01;}
  
  switch (state)
  {
    case BAT_POWER:
      SET_SUPPLY_VOLTAGE_LOW();
      set_state_timer(TIME_DELAY_TO_SLEEP);            
      BAT_OFF();
      MOTOR_OFF();
      led_state = LED_OFF;
      charge_complete = FALSE;
#if DEBUG      
      error_code = NO_ERROR;
#endif      
      PWM_Stop();
      
      if(check_state_timer()) new_state = SLEEP;          
    break;
    
    case SLEEP: 
      SET_SUPPLY_VOLTAGE_LOW();    
      set_state_timer(0); 
    break;    
    
    case BAT_MOTOR:
      SET_SUPPLY_VOLTAGE_HIGH();
      set_state_timer(0); 
      BAT_ON();
      MOTOR_ON();
      led_state = LED_OFF; 
      PWM_Stop();  
      
      if (vload < BAT_VOLT_MOTOR_STOP) new_state = BAT_POWER;
    break;
    
    case EXT_POWER:
      SET_SUPPLY_VOLTAGE_HIGH();  
      set_state_timer(0); 
      BAT_ON();
      MOTOR_OFF();
      led_state = LED_ON;  
      PWM_Start();
      
      prs_val = SUPPLY_CURRENT;
      act_val = iload; 
      limit_val = FALSE; 
                 
      if ((FALSE == charge_complete) || (vload < BAT_VOLT_CHARGE_START)) 
      {
        new_state = CHARGING;
        set_interval_timer(TIME_BAT_VOLT_AND_TEMP);
        prs_val = BAT_CURRENT_TRICLE;
        charge_completion_reset();
      }           
    break;
    
    case EXT_MOTOR:
      SET_SUPPLY_VOLTAGE_HIGH();
      set_state_timer(0); 
      BAT_OFF();
      MOTOR_ON();
      led_state = LED_OFF;  
      PWM_Start();   
       
      prs_val = MOTOR_VOLTAGE_EXT_POW;      
      act_val = vload;                
      limit_val = (iload > MOTOR_MAX_CURRENT);
    break;
    
    case CHARGING:
      SET_SUPPLY_VOLTAGE_HIGH();
      set_state_timer(TIME_TOTAL_CHARGE); 
      MOTOR_OFF();
      BAT_ON();
      PWM_Start();
      led_state = LED_BLINK;
      
      act_val = iload; 
      limit_val = FALSE;
      
      if(check_interval_timer()) 
      {
        set_interval_timer(TIME_BAT_VOLT_AND_TEMP);
        measure_bat_temperature();
        
        if (check_temp_range()) 
        {
          control_state_timer(TRUE);
          prs_val = BAT_CURRENT_RAPID;
        } 
          else
        {
          control_state_timer(FALSE);
          prs_val = BAT_CURRENT_TRICLE;
          charge_completion_reset();
        }
        
        if(charge_completion_check(vload_filt)) 
          if (read_state_timer() < (TIME_TOTAL_CHARGE - TIME_STARTUP_DELAY))
          {
            new_state = EXT_POWER; 
#if DEBUG            
            error_code = CHARGE_COMPLETED_WELL;
#endif            
            charge_complete = TRUE;
          } 
      }
      
      if (check_state_timer())
      { 
        new_state = EXT_POWER;
        charge_complete = TRUE;
#if DEBUG        
        error_code = TIMEOUT_ERROR;
#endif        
      }  
      
      if (vload > BAT_VOLT_OVER)
      {
        new_state = EXT_POWER;
        charge_complete = TRUE;
#if DEBUG        
        error_code = OVER_VOLT_ERROR;
#endif        
      }
    break;
    
    default:
      new_state = BAT_POWER;
    break;  
  }
  
  if(new_state != state) REGULATOR_RESET(); 
  state = new_state;
}

void main()
{
  VLT_CR &= ~0x30;
  
  PRS8_WriteSeed(0x01);
  PRS8_WritePolynomial(0xB8);
  PRS8_Start();
  PRS8_SEED_REG = 0x00;
  
  M8C_EnableIntMask(INT_MSK0, INT_MSK0_SLEEP);  
  M8C_EnableIntMask(INT_MSK0, INT_MSK0_VC3);
    
  M8C_EnableGInt;
  
#if DEBUG  
  Tx232_Start(Tx232_PARITY_NONE);
  SendDebugData();
#endif  
  state = BAT_POWER;
  
  PGA_Start(PGA_MEDPOWER);
  ADCINC_Start(ADCINC_MEDPOWER);
  
  LPF2_Start(LPF2_MEDPOWER);			
  LPF2_FLIN_CR3 |= 0x80;	
  
  while (1)
  {     
    M8C_ClearWDT;  
    
#if DEBUG  
    if (send_debug_info) 
    {
      send_debug_info = FALSE;
      SendDebugData();
    }    
#endif        

     if(SLEEP == state) if(sleep()) continue; 
         
      measure_load_voltage();
      measure_load_current();
      pfm_regulate();          
     
      state_machine();
  }
}
